@@ -1,4 +1,4 @@
-# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
+# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 #
 # SPDX-License-Identifier: Apache-2.0
 
@@ -8,7 +8,7 @@ import ctypes
 import weakref
 from contextlib import contextmanager
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, List, Optional, Tuple, Union
+from typing import TYPE_CHECKING, List, Tuple, Union
 from warnings import warn
 
 if TYPE_CHECKING:
@@ -77,6 +77,7 @@ def _lazy_init():
             "fatbin": _nvjitlink.InputType.FATBIN,
             "ltoir": _nvjitlink.InputType.LTOIR,
             "object": _nvjitlink.InputType.OBJECT,
+            "library": _nvjitlink.InputType.LIBRARY,
         }
     else:
         _driver_input_types = {
@@ -84,6 +85,7 @@ def _lazy_init():
             "cubin": _driver.CUjitInputType.CU_JIT_INPUT_CUBIN,
             "fatbin": _driver.CUjitInputType.CU_JIT_INPUT_FATBINARY,
             "object": _driver.CUjitInputType.CU_JIT_INPUT_OBJECT,
+            "library": _driver.CUjitInputType.CU_JIT_INPUT_LIBRARY,
         }
     _inited = True
 
@@ -98,6 +100,8 @@ class LinkerOptions:
 
     Attributes
     ----------
+    name : str, optional
+        Name of the linker. If the linking succeeds, the name is passed down to the generated `ObjectCode`.
     arch : str, optional
         Pass the SM architecture value, such as ``sm_<CC>`` (for generating CUBIN) or
         ``compute_<CC>`` (for generating PTX). If not provided, the current device's architecture
@@ -159,29 +163,31 @@ class LinkerOptions:
         Default: False.
     """
 
-    arch: Optional[str] = None
-    max_register_count: Optional[int] = None
-    time: Optional[bool] = None
-    verbose: Optional[bool] = None
-    link_time_optimization: Optional[bool] = None
-    ptx: Optional[bool] = None
-    optimization_level: Optional[int] = None
-    debug: Optional[bool] = None
-    lineinfo: Optional[bool] = None
-    ftz: Optional[bool] = None
-    prec_div: Optional[bool] = None
-    prec_sqrt: Optional[bool] = None
-    fma: Optional[bool] = None
-    kernels_used: Optional[Union[str, Tuple[str], List[str]]] = None
-    variables_used: Optional[Union[str, Tuple[str], List[str]]] = None
-    optimize_unused_variables: Optional[bool] = None
-    ptxas_options: Optional[Union[str, Tuple[str], List[str]]] = None
-    split_compile: Optional[int] = None
-    split_compile_extended: Optional[int] = None
-    no_cache: Optional[bool] = None
+    name: str | None = "<default linker>"
+    arch: str | None = None
+    max_register_count: int | None = None
+    time: bool | None = None
+    verbose: bool | None = None
+    link_time_optimization: bool | None = None
+    ptx: bool | None = None
+    optimization_level: int | None = None
+    debug: bool | None = None
+    lineinfo: bool | None = None
+    ftz: bool | None = None
+    prec_div: bool | None = None
+    prec_sqrt: bool | None = None
+    fma: bool | None = None
+    kernels_used: Union[str, Tuple[str], List[str]] | None = None
+    variables_used: Union[str, Tuple[str], List[str]] | None = None
+    optimize_unused_variables: bool | None = None
+    ptxas_options: Union[str, Tuple[str], List[str]] | None = None
+    split_compile: int | None = None
+    split_compile_extended: int | None = None
+    no_cache: bool | None = None
 
     def __post_init__(self):
         _lazy_init()
+        self._name = self.name.encode()
         self.formatted_options = []
         if _nvjitlink:
             self._init_nvjitlink()
@@ -391,7 +397,7 @@ class Linker:
         data = object_code._module
         assert_type(data, bytes)
         with _exception_manager(self):
-            name_str = f"{object_code._handle}_{object_code._code_type}"
+            name_str = f"{object_code.name}"
             if _nvjitlink:
                 _nvjitlink.add_data(
                     self._mnff.handle,
@@ -453,7 +459,7 @@ class Linker:
                 addr, size = handle_return(_driver.cuLinkComplete(self._mnff.handle))
                 code = (ctypes.c_char * size).from_address(addr)
 
-        return ObjectCode._init(bytes(code), target_type)
+        return ObjectCode._init(bytes(code), target_type, name=self._options.name)
 
     def get_error_log(self) -> str:
         """Get the error log generated by the linker.
