@@ -1,4 +1,4 @@
-# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
+# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 #
 # SPDX-License-Identifier: Apache-2.0
 
@@ -6,20 +6,21 @@ from __future__ import annotations
 
 import abc
 import weakref
-from typing import Optional, Tuple, TypeVar, Union
+from typing import Tuple, TypeVar, Union
 
 from cuda.core.experimental._dlpack import DLDeviceType, make_py_capsule
-from cuda.core.experimental._stream import default_stream
+from cuda.core.experimental._stream import Stream, default_stream
 from cuda.core.experimental._utils.cuda_utils import driver, handle_return
 
-PyCapsule = TypeVar("PyCapsule")
-
-
 # TODO: define a memory property mixin class and make Buffer and
 # MemoryResource both inherit from it
 
+
+PyCapsule = TypeVar("PyCapsule")
+"""Represent the capsule type."""
+
 DevicePointerT = Union[driver.CUdeviceptr, int, None]
-"""A type union of `Cudeviceptr`, `int` and `None` for hinting Buffer.handle."""
+"""A type union of :obj:`~driver.CUdeviceptr`, `int` and `None` for hinting :attr:`Buffer.handle`."""
 
 
 class Buffer:
@@ -29,19 +30,7 @@ class Buffer:
     different memory resources are to give access to their memory
     allocations.
 
-    Support for data interchange mechanisms are provided by
-    establishing both the DLPack and the Python-level buffer
-    protocols.
-
-    Parameters
-    ----------
-    ptr : Any
-        Allocated buffer handle object
-    size : Any
-        Memory size of the buffer
-    mr : :obj:`~_memory.MemoryResource`, optional
-        Memory resource associated with the buffer
-
+    Support for data interchange mechanisms are provided by DLPack.
     """
 
     class _MembersNeededForFinalize:
@@ -55,8 +44,6 @@ class Buffer:
 
         def close(self, stream=None):
             if self.ptr and self.mr is not None:
-                if stream is None:
-                    stream = default_stream()
                 self.mr.deallocate(self.ptr, self.size, stream)
                 self.ptr = 0
                 self.mr = None
@@ -64,10 +51,16 @@ class Buffer:
     # TODO: handle ownership? (_mr could be None)
     __slots__ = ("__weakref__", "_mnff")
 
-    def __init__(self, ptr, size, mr: MemoryResource = None):
-        self._mnff = Buffer._MembersNeededForFinalize(self, ptr, size, mr)
+    def __new__(self, *args, **kwargs):
+        raise RuntimeError("Buffer objects cannot be instantiated directly. Please use MemoryResource APIs.")
 
-    def close(self, stream=None):
+    @classmethod
+    def _init(cls, ptr: DevicePointerT, size: int, mr: MemoryResource | None = None):
+        self = super().__new__(cls)
+        self._mnff = Buffer._MembersNeededForFinalize(self, ptr, size, mr)
+        return self
+
+    def close(self, stream: Stream = None):
         """Deallocate this buffer asynchronously on the given stream.
 
         This buffer is released back to their memory resource
@@ -75,11 +68,9 @@ class Buffer:
 
         Parameters
         ----------
-        stream : Any, optional
-            The stream object with a __cuda_stream__ protocol to
-            use for asynchronous deallocation. Defaults to using
-            the default stream.
-
+        stream : Stream, optional
+            The stream object to use for asynchronous deallocation. If None,
+            the behavior depends on the underlying memory resource.
         """
         self._mnff.close(stream)
 
@@ -95,7 +86,7 @@ class Buffer:
         return self._mnff.ptr
 
     @property
-    def size(self):
+    def size(self) -> int:
         """Return the memory size of this buffer."""
         return self._mnff.size
 
@@ -125,7 +116,7 @@ class Buffer:
             return self._mnff.mr.device_id
         raise NotImplementedError("WIP: Currently this property only supports buffers with associated MemoryResource")
 
-    def copy_to(self, dst: Buffer = None, *, stream) -> Buffer:
+    def copy_to(self, dst: Buffer = None, *, stream: Stream) -> Buffer:
         """Copy from this buffer to the dst buffer asynchronously on the given stream.
 
         Copies the data from this buffer to the provided dst buffer.
@@ -136,7 +127,7 @@ class Buffer:
         ----------
         dst : :obj:`~_memory.Buffer`
             Source buffer to copy data from
-        stream : Any
+        stream : Stream
             Keyword argument specifying the stream for the
             asynchronous copy
 
@@ -154,14 +145,14 @@ class Buffer:
         handle_return(driver.cuMemcpyAsync(dst._mnff.ptr, self._mnff.ptr, self._mnff.size, stream.handle))
         return dst
 
-    def copy_from(self, src: Buffer, *, stream):
+    def copy_from(self, src: Buffer, *, stream: Stream):
         """Copy from the src buffer to this buffer asynchronously on the given stream.
 
         Parameters
         ----------
         src : :obj:`~_memory.Buffer`
             Source buffer to copy data from
-        stream : Any
+        stream : Stream
             Keyword argument specifying the stream for the
             asynchronous copy
 
@@ -177,10 +168,10 @@ class Buffer:
     def __dlpack__(
         self,
         *,
-        stream: Optional[int] = None,
-        max_version: Optional[Tuple[int, int]] = None,
-        dl_device: Optional[Tuple[int, int]] = None,
-        copy: Optional[bool] = None,
+        stream: int | None = None,
+        max_version: Tuple[int, int] | None = None,
+        dl_device: Tuple[int, int] | None = None,
+        copy: bool | None = None,
     ) -> PyCapsule:
         # Note: we ignore the stream argument entirely (as if it is -1).
         # It is the user's responsibility to maintain stream order.
@@ -219,107 +210,247 @@ class Buffer:
         # Supporting method paired with __buffer__.
         raise NotImplementedError("WIP: Buffer.__release_buffer__ hasn't been implemented yet.")
 
+    @staticmethod
+    def from_handle(ptr: DevicePointerT, size: int, mr: MemoryResource | None = None) -> Buffer:
+        """Create a new :class:`Buffer` object from a pointer.
+
+        Parameters
+        ----------
+        ptr : :obj:`~_memory.DevicePointerT`
+            Allocated buffer handle object
+        size : int
+            Memory size of the buffer
+        mr : :obj:`~_memory.MemoryResource`, optional
+            Memory resource associated with the buffer
+        """
+        return Buffer._init(ptr, size, mr=mr)
+
 
 class MemoryResource(abc.ABC):
+    """Abstract base class for memory resources that manage allocation and deallocation of buffers.
+
+    Subclasses must implement methods for allocating and deallocation, as well as properties
+    associated with this memory resource from which all allocated buffers will inherit. (Since
+    all :class:`Buffer` instances allocated and returned by the :meth:`allocate` method would
+    hold a reference to self, the buffer properties are retrieved simply by looking up the underlying
+    memory resource's respective property.)
+    """
+
     __slots__ = ("_handle",)
 
     @abc.abstractmethod
-    def __init__(self, *args, **kwargs): ...
+    def __init__(self, *args, **kwargs):
+        """Initialize the memory resource.
+
+        Subclasses may use additional arguments to configure the resource.
+        """
+        ...
 
     @abc.abstractmethod
-    def allocate(self, size, stream=None) -> Buffer: ...
+    def allocate(self, size: int, stream: Stream = None) -> Buffer:
+        """Allocate a buffer of the requested size.
+
+        Parameters
+        ----------
+        size : int
+            The size of the buffer to allocate, in bytes.
+        stream : Stream, optional
+            The stream on which to perform the allocation asynchronously.
+            If None, it is up to each memory resource implementation to decide
+            and document the behavior.
+
+        Returns
+        -------
+        Buffer
+            The allocated buffer object, which can be used for device or host operations
+            depending on the resource's properties.
+        """
+        ...
 
     @abc.abstractmethod
-    def deallocate(self, ptr, size, stream=None): ...
+    def deallocate(self, ptr: DevicePointerT, size: int, stream: Stream = None):
+        """Deallocate a buffer previously allocated by this resource.
+
+        Parameters
+        ----------
+        ptr : :obj:`~_memory.DevicePointerT`
+            The pointer or handle to the buffer to deallocate.
+        size : int
+            The size of the buffer to deallocate, in bytes.
+        stream : Stream, optional
+            The stream on which to perform the deallocation asynchronously.
+            If None, it is up to each memory resource implementation to decide
+            and document the behavior.
+        """
+        ...
 
     @property
     @abc.abstractmethod
     def is_device_accessible(self) -> bool:
-        # Check if the buffers allocated from this MR can be accessed from
-        # GPUs.
+        """bool: True if buffers allocated by this resource can be accessed on the device."""
         ...
 
     @property
     @abc.abstractmethod
     def is_host_accessible(self) -> bool:
-        # Check if the buffers allocated from this MR can be accessed from
-        # CPUs.
+        """bool: True if buffers allocated by this resource can be accessed on the host."""
         ...
 
     @property
     @abc.abstractmethod
     def device_id(self) -> int:
-        # Return the device ID if this MR is for single devices. Raise an
-        # exception if it is not.
+        """int: The device ordinal for which this memory resource is responsible.
+
+        Raises
+        ------
+        RuntimeError
+            If the resource is not bound to a specific device.
+        """
         ...
 
 
-class _DefaultAsyncMempool(MemoryResource):
+class DeviceMemoryResource(MemoryResource):
+    """Create a device memory resource that uses the driver's stream-ordered memory pool.
+
+    Parameters
+    ----------
+    device_id : int
+        Device ordinal for which a memory resource is constructed. The mempool that is
+        set to *current* on ``device_id`` is used. If no mempool is set to current yet,
+        the driver would use the *default* mempool on the device.
+    """
+
     __slots__ = ("_dev_id",)
 
-    def __init__(self, dev_id):
-        self._handle = handle_return(driver.cuDeviceGetMemPool(dev_id))
-        self._dev_id = dev_id
+    def __init__(self, device_id: int):
+        self._handle = handle_return(driver.cuDeviceGetMemPool(device_id))
+        self._dev_id = device_id
 
-    def allocate(self, size, stream=None) -> Buffer:
+    def allocate(self, size: int, stream: Stream = None) -> Buffer:
+        """Allocate a buffer of the requested size.
+
+        Parameters
+        ----------
+        size : int
+            The size of the buffer to allocate, in bytes.
+        stream : Stream, optional
+            The stream on which to perform the allocation asynchronously.
+            If None, an internal stream is used.
+
+        Returns
+        -------
+        Buffer
+            The allocated buffer object, which is accessible on the device that this memory
+            resource was created for.
+        """
         if stream is None:
             stream = default_stream()
         ptr = handle_return(driver.cuMemAllocFromPoolAsync(size, self._handle, stream.handle))
-        return Buffer(ptr, size, self)
+        return Buffer._init(ptr, size, self)
 
-    def deallocate(self, ptr, size, stream=None):
+    def deallocate(self, ptr: DevicePointerT, size: int, stream: Stream = None):
+        """Deallocate a buffer previously allocated by this resource.
+
+        Parameters
+        ----------
+        ptr : :obj:`~_memory.DevicePointerT`
+            The pointer or handle to the buffer to deallocate.
+        size : int
+            The size of the buffer to deallocate, in bytes.
+        stream : Stream, optional
+            The stream on which to perform the deallocation asynchronously.
+            If None, an internal stream is used.
+        """
         if stream is None:
             stream = default_stream()
         handle_return(driver.cuMemFreeAsync(ptr, stream.handle))
 
     @property
     def is_device_accessible(self) -> bool:
+        """bool: this memory resource provides device-accessible buffers."""
         return True
 
     @property
     def is_host_accessible(self) -> bool:
+        """bool: this memory resource does not provides host-accessible buffers."""
         return False
 
     @property
     def device_id(self) -> int:
+        """int: the associated device ordinal."""
         return self._dev_id
 
 
-class _DefaultPinnedMemorySource(MemoryResource):
+class LegacyPinnedMemoryResource(MemoryResource):
+    """Create a pinned memory resource that uses legacy cuMemAllocHost/cudaMallocHost
+    APIs.
+    """
+
     def __init__(self):
         # TODO: support flags from cuMemHostAlloc?
         self._handle = None
 
-    def allocate(self, size, stream=None) -> Buffer:
-        ptr = handle_return(driver.cuMemAllocHost(size))
-        return Buffer(ptr, size, self)
+    def allocate(self, size: int, stream: Stream = None) -> Buffer:
+        """Allocate a buffer of the requested size.
 
-    def deallocate(self, ptr, size, stream=None):
+        Parameters
+        ----------
+        size : int
+            The size of the buffer to allocate, in bytes.
+        stream : Stream, optional
+            Currently ignored
+
+        Returns
+        -------
+        Buffer
+            The allocated buffer object, which is accessible on both host and device.
+        """
+        ptr = handle_return(driver.cuMemAllocHost(size))
+        return Buffer._init(ptr, size, self)
+
+    def deallocate(self, ptr: DevicePointerT, size: int, stream: Stream = None):
+        """Deallocate a buffer previously allocated by this resource.
+
+        Parameters
+        ----------
+        ptr : :obj:`~_memory.DevicePointerT`
+            The pointer or handle to the buffer to deallocate.
+        size : int
+            The size of the buffer to deallocate, in bytes.
+        stream : Stream, optional
+            The stream on which to perform the deallocation asynchronously.
+            If None, no synchronization would happen.
+        """
+        if stream:
+            stream.sync()
         handle_return(driver.cuMemFreeHost(ptr))
 
     @property
     def is_device_accessible(self) -> bool:
+        """bool: this memory resource provides device-accessible buffers."""
         return True
 
     @property
     def is_host_accessible(self) -> bool:
+        """bool: this memory resource provides host-accessible buffers."""
         return True
 
     @property
     def device_id(self) -> int:
+        """This memory resource is not bound to any GPU."""
         raise RuntimeError("a pinned memory resource is not bound to any GPU")
 
 
 class _SynchronousMemoryResource(MemoryResource):
     __slots__ = ("_dev_id",)
 
-    def __init__(self, dev_id):
+    def __init__(self, device_id):
         self._handle = None
-        self._dev_id = dev_id
+        self._dev_id = device_id
 
     def allocate(self, size, stream=None) -> Buffer:
         ptr = handle_return(driver.cuMemAlloc(size))
-        return Buffer(ptr, size, self)
+        return Buffer._init(ptr, size, self)
 
     def deallocate(self, ptr, size, stream=None):
         if stream is None:
